// Tree walker for ANTLr pretty printer.




header {
/* ANTLR grammar tree walker
 * ANTLRBase Copyright 2006 Warp IV Technologies
 * Author:  Loring Craymer
 */

import antlr.StringUtils;
import java.io.*;
import antlr.collections.AST;
}

class ANTLRBaseTree extends TreeParser;
options {
	buildAST = false;
	importVocab = ANTLRBase;
}


{	boolean acting = true;
	boolean opting = true;
	boolean commenting = true;
	boolean documenting = true;
	boolean usingArgs = true;
	boolean excepting = true;
	boolean throwing = true;

	private FileOutputStream _out = null;
	private PrintStream p = null;
	private File file = null;
	protected int inset = 0;
	protected String path = null;
	protected String fileName = null;


	public void setActing(boolean act) {
		acting = act;
	}

	public void setOpting(boolean opt) {
		opting = opt;
	}

	public void setCommenting(boolean cmt) {
		commenting = cmt;
	}

	public void setExcepting(boolean exc) {
		excepting = exc;
	}

	public void setThrowing(boolean thr) {
		throwing = thr;
	}

	public void setDocumenting(boolean doc) {
		documenting = doc;
	}

	public void setUsingArgs(boolean args) {
		usingArgs = args;
	}

	public void setPath(String path_) {
		path = path_;
	}

	public void setFileName(String name) {
		fileName = name;
	}

	public void bracket(String label) {
		String s = tabs(inset) + label + "(";
		inset++;
		println(s);
	}

	public void endBracket(String type) {
		inset--;
		String s = null;
		if (type != null)
			s = tabs(inset) + type;
		else
			s = tabs(inset) + ")";

		println(s);
	}

	public void startTree(AST ast) {
		String s = tabs(inset) + "#( ";
		println(s);
		inset++;
		print(ast);
	}

	public void startTree(String text) {
		String s = tabs(inset) + "#( ";
		println(s);
		inset++;
		print(text);
	}

	public void startTypeTree(AST ast) {
		String s = tabs(inset) + "#( ";
		println(s);
		inset++;
		printType(ast);
	}


	public void startRule(AST ast) {
		inset = 0;
		println("");
		String s = ast.getText();
		println(s);
		inset++;
		println(tabs(inset) + ":");
	}

	public void endRule() {
		println("\t;");
	}

	public void print(String str) {
		String s = tabs(inset) + str;
		println(s);
	}
	
	public void print(AST ast) {
		String s = tabs(inset) + ast.getText();
		println(s);
	}


	public void printGrammarName(String type, AST ident, String sup) {
		String text = "class " + ident.getText() + "extends ";

		text += type;
		if (sup != null)
			text += "(" + sup + ")";

		text += ";";

		print(text);
	}

	// FIXME -- should pretty print action
	public void printAction(AST ast) {
		if (acting) {
			String s = tabs(inset) + ast.getText();
			println(s);
		}
	}

	public void printHandler(AST arg, AST action) {
		if (excepting) {
			print("catch " + #arg.getText());
			if (acting) {
				String s = tabs(inset) + action.getText();
				println(s);
			}
			else
				println(tabs(inset) + "{}");
		}
	}

	public void printDoc(AST ast) {
		if (documenting)
			print(ast);
	}

	public void printOption(AST which, String value) {
		String option = which.getText() + " = " + value + ";";
		if (opting) {
			print(option);
		}
	}

	public void printType(AST ast) {
		String astCtorArgs = ast.getText();
		if (astCtorArgs == null) {
			astCtorArgs = "";
		}

		int nCommas = 0;

		for (int i = 0; i < astCtorArgs.length(); i++) {
			if (astCtorArgs.charAt(i) == ',') {
				nCommas++;
			}
		}

		//System.out.println("num commas="+nCommas);
		int firstComma = astCtorArgs.indexOf(',');
		int lastComma = astCtorArgs.lastIndexOf(',');
		String typeString = astCtorArgs;

		if (nCommas > 0) {
			typeString = astCtorArgs.substring(0, firstComma);
		}

		println(tabs(inset) + typeString);
	}
		

	public void initOut(String name) {
		String fileName = name;

		try {
			file = new File(path, fileName);
			_out = new FileOutputStream(file);
		}
		catch (FileNotFoundException e) {
			System.out.println("file not found");
			System.exit(1);
		}
		catch (IOException ie) {
			System.out.println("I/O exception");
			System.exit(1);
		}
		p = new PrintStream(_out);
		printHeader();
	}

	protected void printHeader() {
		p.println("// ANTLR grammar");
		p.println("\n\n\n");
		p.println("header {");
		p.println("}");
		p.println("");
	}

	protected String tabs(int n) {
		String result = "";
		for (int i = 0; i<n; i++)
			result += "\t";
	
		return result;
	}

	protected void println(String s) {
		p.println(s);
	}	
}


grammar
	:
	comments
	headerSection
	(
		optionsSpec
	)?
	(
		classDef
	)*
	EOF
	;


headerSection
{	String name = "";
}
	:
	#( 
		LITERAL_header
		headerSection
		(
			nm:STRING_LITERAL
			{ name = #nm.getText(); }
		)?
		a:ACTION
		{	print( "header " + name + " ");
			printAction(#a);
		}
		comments
	)
	|
	;

comments
	:
	(	options { greedy = true; } :
		c:SL_COMMENT
		{ print(#c); }
		|
			m:ML_COMMENT
			{ print(#m); }
	)*
	;

classDef
	:
	(
		act:ACTION
		{ printAction(#act); }
	)?
	(
		doc:DOC_COMMENT
		{ printDoc(#doc); }
	)?
	comments
	(
		lexerSpec
		|
			treeParserSpec
		|
			parserSpec
	)
	rules
	;

optionsSpec
	:
	#( 
		RCURLY
		OPTIONS
		{	if (opting) {
				println("options {");
				inset++;
			}
		}
		(
			option
		)*
		{	if (opting) {
				inset--;
				println("}");
			}
		}
	)
	;

option
{ String value = null; }
	:
	#( 
		ASSIGN
		i:id
		value = optionValue
		{ printOption(#i, value); }
		comments
	)
	;


optionString
returns [ String opt = null ]
{ String value = null; }
	:
	#( 
		ASSIGN
		i:id
		value = optionValue
		comments
		{ opt = #i.getText() + " = " + value; }
	)
	;


lexerOptionsSpec
	:
	#( 
		RCURLY
		OPTIONS
		{	if (opting) {
				println("options {");
				inset++;
			}
		}
		(
			lexerOption
		)*
		{	if (opting) {
				inset--;
				println("}");
			}
		}
	)
	;

lexerOption
{	String setValue = null;
	String value = null;
}
	:
	#( 
		ASSIGN
		(
			"charVocabulary"
			setValue = charSet
			{	if (opting)
					print("charVocabulary = " + setValue);
			}
		|
			i:id
			value = optionValue
			{ printOption(#i, value); }
		)
		comments
	)
	;

optionValue
returns [ String text = null ]
	:
	text = qualifiedID
	|
		s:STRING_LITERAL	{ text = #s.getText(); }
	|
		c:CHAR_LITERAL		{ text = #c.getText(); }
	|
		i:INT			{ text = #i.getText(); }
	;


charSet
returns [ String setValue=null ]
{	String blockVal = null;
}
	:
	#( 
		OR
		setValue = charSet
		blockVal = setBlockElement
		{ setValue += " | " + blockVal; }
	)
	|
		setValue = setBlockElement
	;

setBlockElement
returns [String blockValue = null ]
	:
	#( 
		RANGE
		c1:CHAR_LITERAL
		c2:CHAR_LITERAL
		{ blockValue = #c1.getText() + ".." + #c2.getText(); }
	)
	|
		c3:CHAR_LITERAL
		{ blockValue = #c3.getText(); }
	;

tokensSpec
	:
	#( 
		TOKENS
		{	println("tokens {");
			inset++;
		}
		(
			tokenSpecItem
		)+
		{	inset--;
			println("}");
		}
	)
	;

tokenSpecItem
{ String text = null; String ret = null; }
	:
	#( 
		SEMI
		(
			#( 
				ASSIGN
				t:TOKEN_REF
				s:STRING_LITERAL
				{ text = #t.getText() + " = " + #s.getText(); }
			)
			|
				to:TOKEN_REF
				{ text = #to.getText(); }
			|
				sl:STRING_LITERAL
				{ text = #sl.getText(); }
		)
		(
			ret = tokensSpecOptions
			{ text += "<" + ret + ">"; }
		)?
		{ print(text + ";"); }
		comments
	)
	;

tokensSpecOptions
returns [ String text = null ]
{ String ret = null; }
	:
	#( 
		CLOSE_ELEMENT_OPTION
		(
			ret = optionString
			{	if (text == null)
					text = ret;
				else
					text += ", " + ret;
			}
		)+
	)
	;

superClass
returns [ String parent = null ]
	:
	#( 
		RPAREN
		(
			s:STRING_LITERAL { parent = #s.getText(); }
		)
	)
	;

parserSpec
{ String sup = null; }
	:
	#( 
		"Parser"
		i:id
		(
			sup = superClass
		)?
		{ printGrammarName("Parser", #i, sup); }
		comments
		(
			optionsSpec
		)?
		(
			tokensSpec
		)?
		(
			act:ACTION
			{ printAction(#act); }
		)?
	)
	;

lexerSpec
{ String sup = null; }
	:
	#( 
		"lexclass"
		i:id
		{ printGrammarName("Lexer", #i, null); }
		comments
		(
			lexerOptionsSpec
		)?
		(
			tokensSpec
		)?
		(
			act:ACTION
			{ printAction(#act); }
		)?
	)
	|
		#( 
			"Lexer"
			i2:id
			(
				sup = superClass
			)?
			{ printGrammarName("Lexer", #i2, sup); }

			comments
			(
				lexerOptionsSpec
			)?
			(
				tokensSpec
			)?
			(
				ac:ACTION
				{ printAction(#ac); }
			)?
		)
	;

treeParserSpec
{ String sup = null; }
	:
	#( 
		"TreeParser"
		i:id
		(
			sup = superClass
		)?
		{ printGrammarName("TreeParser", #i, sup); }
		comments
		(
			optionsSpec
		)?
		(
			tokensSpec
		)?
		(
			act:ACTION
			{ printAction(#act); }
		)?
	)
	;

rules
	:
	(
		rule
	)+
	;

rule
{	String prefix = "";
	String in = null;
	boolean forget = false;
}
	:
	{ inset = 0; }
	#( 
		COLON
		comments
		(
			doc:DOC_COMMENT
			{ printDoc(#doc); }
			comments
		)?
		(
			"protected"	{ prefix = "protected"; }
			|
				"public"	{ prefix = "public"; }
			|
				"private"	{ prefix = "private"; }
		)?
		ruleId:id
		(
			BANG	{ forget = true; }
		)?
		(
			arg:ARG_ACTION	{ in = #arg.getText(); }
		)?
		comments
		{	if (prefix != null)
				print(prefix);

			if (usingArgs && in != null)
				print(#ruleId.getText() + in);
			else
				print(#ruleId);

		}
		(
			"returns"
			arg2:ARG_ACTION
			{ print("returns " + #arg2.getText()); }
			comments
		)?
		(
			throwsSpec
		)?
		(
			optionsSpec
		)?
		(
			r:RULE_ACTION
			{ printAction(#r); }
		)?
		{ inset++; print(":"); }
		block
		SEMI
		{ print(";"); inset--; }
		(
			exceptionGroup
		)?
	)
	;

throwsSpec
{ String targets = null; }
	:
	#( 
		"throws"
		i:id		{ targets = #i.getText(); }
		(
			i2:id	{ targets += ", " + #i2.getText(); }
		)*
	)
	{	if (throwing)
			print("throws " + targets);
	}
	;


block
{ boolean bang = false; }
	:
	(
		NO_ALT
		{ print("!"); }
	)?
	alternative 
	(	OR
		{ print("|"); inset++; }
		alternative
		(	OR
			(
				{ bang = true; }
				NO_ALT
			|
				{ bang = false; }
			)
			{	inset--;
				if (bang)
					print("|!");
				else
					print("|");
				inset++; }
			comments
			alternative
		)*
		{ inset--; }
	)?
	;

alternative
	:
	comments
	(
		options { greedy = true; } :
		element
	)*
	(
		exceptionSpec
	)?
	;

exceptionGroup
	:
	(
		exceptionSpec
	)+
	;

exceptionSpec
	:
	#( 
		"exception"
		(
			arg:ARG_ACTION
			{ print("exception" + #arg.getText()); }
		|
			{ print("exception"); }
		)
		(
			exceptionHandler
		)*
	)
	;

exceptionHandler
	:
	#( 
		"catch"
		arg:ARG_ACTION
		act:ACTION
		{ printHandler(#arg, #act); }
	)
	;

element
	:
	elementNoOptionSpec
	comments
	(
		elementOptionSpec
	)?
	;

elementOptionSpec
	:
	#( 
		CLOSE_ELEMENT_OPTION
		{	if (opting) {
				print("option {");
				inset++;
			}
		}
		( option )+
		{	if (opting) {
				inset--;
				print("}");
			}
		}
	)
	;


elementNoOptionSpec
{	String text = null;
	String ret = null;
}
	:
	#( 
		COLON
		(
			(	i:id	{ text = #i.getText() + ":"; }
				ret = longElement[text]
			)
			{ if (ret != null) print(text + ret); }
			comments
		|
			(
				(
					#( 
						ASSIGN
						i1:id
						i2:id
						{ text = #i1.getText() + " - ";
						  text += #i2.getText() + ":";
						}
					)
					(	r:RULE_REF
						{ text += #r.getText(); }
						(
							arg:ARG_ACTION
							{ text += #arg.getText(); }
	
						)?
						(
							BANG
							{ text += "!"; }
						)?
						|
							t:TOKEN_REF
							{ text += #t.getText(); }
							(
								arg2:ARG_ACTION
								{ text += #arg.getText(); }
							)?
					)
				)
				{ print(text); }
				comments
			)
		)
	)
	|
		text = longElement[""]
		{ if (text != null) print(text); }
		comments
	|
		act:ACTION
		{ printAction(#act); }
		comments
	|
		sem:SEMPRED
		{ print(#sem); }
		comments
	|
		tree
		comments
	|
		#( 
			ASSIGN
			(
				ida:id
				{ text = #ida.getText() + "="; }
				(
					ra:RULE_REF
					{ text += #ra.getText(); }
					(
						arga:ARG_ACTION
						{ text += #arga.getText(); }
					)?
					(
						BANG
						{ text += "!"; }
					)?
					|
						ta:TOKEN_REF
						{ text += #ta.getText(); }
						(
							argb:ARG_ACTION
							{ text += #argb.getText(); }
						)?
				)
				{ print(text); }
			)
			comments
		)
	;


longElement[String label]
returns [String text = null ]
{ String ret = null; }
	:
	{	if (label != null)
			text = label;
		else
			text = "";
	}
	(
		r:RULE_REF
			{ text += #r.getText(); }
		(
			arg:ARG_ACTION
			{ text += #arg.getText(); }
		)?
		(
			BANG
			{ text += "!"; }
		)?
		|
			ret = range
			{ text += ret; }
		|
			text = terminal
		|
			NOT_OP	{	if (label != null)
						label += "~";
					else
						label = "~";
					text = label;
				}
			(
				ret = notTerminal
				{ text += ret; }
				|
					ebnf[label]
					{ text = null; }
			)
		|
			ebnf[label]
			{ text = null; }
	)
	;

tree
{ String base = null; }
	:
	#( 
		TREE_BEGIN
		base = rootNode
		{ startTree(base); }
		(
			element
		)+
		{ endBracket(""); }
	)
	;

rootNode
returns [String ret = null ]
{ String val = null; }
	:
	#( 
		COLON
		i:id
		val = terminal
		{ ret = #i.getText() + ":" + val; }
		comments
	)
	|
		ret = terminal
		comments
	;

ebnf[String label]
{ String text = null; }
	:
	#( 
		LPAREN
		{ bracket(label); }
		comments
		(
			optionsSpec
			(
				act:ACTION
				{ printAction(#act); }
			)?
			COLON
			{ print(":"); }
			|
				( ACTION COLON )=>
				act2:ACTION
				{ printAction(#act2); }
				COLON
				{ print(":"); }
		)?
		block
		RPAREN
		(
			(
				QUESTION	{ text = ")?"; }
				|
					STAR	{ text = ")*"; }
				|
					PLUS	{ text = ")+"; }
			)?
			(
				BANG		{ text += "!"; }
			)?
			|
				IMPLIES		{ text = ")=>"; }
		)
		{ endBracket(text); }
		comments
	)
	;

range
returns [String value = null ]
	:
	#( 
		RANGE
		(
			(
				(
					t:TOKEN_REF
					{ value = #t.getText(); }
					|
						s:STRING_LITERAL
						{ value = #s.getText(); }
				)
				(	t1:TOKEN_REF
					{ value += " .. " + #t1.getText(); }
					|
						s1:STRING_LITERAL
						{ value += " .. " + #s1.getText(); }
				)
			)
			(	CARET		{ value += "^"; }
				|
					BANG	{ value += "!"; }
			)?
			|
				c:CHAR_LITERAL
				{ value = #c.getText(); }
				c1:CHAR_LITERAL
				{ value += ".." + #c1.getText(); }
				(
					BANG	{ value += "!"; }
				)?
		)
	)
	;

terminal
returns [String value = null ]
	:
	c:CHAR_LITERAL	{ value = #c.getText(); }
	(
		BANG
		{ value += "!"; }
	)?
	|
		t:TOKEN_REF		{ value = #t.getText(); }
		(	CARET		{ value += "^"; }
			|
				BANG	{ value += "!"; }
		)?
		(
			ARG_ACTION
		)?
	|
		s:STRING_LITERAL	{ value = #s.getText(); }
		(	CARET		{ value += "^"; }
			|
				BANG	{ value += "!"; }
		)?
	|
		w:WILDCARD		{ value = #w.getText(); }
		(	CARET		{ value += "^"; }
			|
				BANG	{ value += "!"; }
		)?
	;

notTerminal
returns [ String value = null ]
	:
	c:CHAR_LITERAL	{ value = #c.getText(); }
	(
		BANG
	)?
	|
		t:TOKEN_REF		{ value = #t.getText(); }
		(	CARET		{ value += "^"; }
			|
				BANG	{ value += "!"; }
		)?
	;

qualifiedID
returns [ String text = null ]
	:
	#( 
		WILDCARD
		text = qualifiedID
		i:id			{ text += "." + #i.getText(); }
	)
	|
		i1:id			{ text = #i1.getText(); }
	;

id
	:
	TOKEN_REF
	|
		RULE_REF
	;
